import math
import time
import numpy as np
import multiprocessing as mp

from typing import List, Callable

#######################################
##### KY MATRIX FUNCTIONS
#######################################

def get_ky_mat(hyps: np.ndarray, training_data: list,
               kernel, cutoffs=None, hyps_mask=None):

    """ Compute covariance matrix K by comparing training data with itself
    :param hyps: list of hyper-parameters
    :param training_data: list of atomic envirionments
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """

    # assume sigma_n is the final hyperparameter
    number_of_hyps = len(hyps)
    sigma_n = hyps[number_of_hyps - 1]
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask):
            if (hyps_mask['train_noise'] is False):
                sigma_n = hyps_mask['original'][-1]

    # matrix manipulation
    k_mat = get_ky_mat_pack(hyps, training_data,
                            training_data, True, kernel,
                            cutoffs, hyps_mask)
    size3 = len(training_data)*3
    ky_mat = k_mat + sigma_n ** 2 * np.eye(size3)

    return ky_mat


def get_ky_mat_pack(hyps: np.ndarray, training_data1: list,
               training_data2:list, same: bool,
               kernel, cutoffs=None, hyps_mask=None):
    """ Compute covariance matrix element between set1 and set2
    :param hyps: list of hyper-parameters
    :param training_data1: row elements
    :param training_data2: column elements
    :param same: whether the row and column are the same
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """


    # initialize matrices
    size1 = len(training_data1) * 3
    size2 = len(training_data2) * 3
    k_mat = np.zeros([size1, size2])

    ds = [1, 2, 3]

    # calculate elements
    for m_index in range(size1):
        x_1 = training_data1[int(math.floor(m_index / 3))]
        d_1 = ds[m_index % 3]
        if (same):
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data2[int(math.floor(n_index / 3))]
            d_2 = ds[n_index % 3]

            # calculate kernel and gradient
            if (hyps_mask is not None):
                kern_curr = kernel(x_1, x_2, d_1, d_2, hyps,
                                   cutoffs, hyps_mask=hyps_mask)
            else:
                kern_curr = kernel(x_1, x_2, d_1, d_2, hyps,
                                   cutoffs)

            # store kernel value
            k_mat[m_index, n_index] = kern_curr
            if (same):
                k_mat[n_index, m_index] = kern_curr

    return k_mat


def get_ky_mat_par(hyps: np.ndarray, training_data: list,
                   kernel, cutoffs=None, hyps_mask=None,
                   n_cpus=None, nsample=100):
    """ parallel version of get_ky_mat
    :param hyps: list of hyper-parameters
    :param training_data: list of atomic envirionments
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """

    if (n_cpus is None):
        n_cpus =mp.cpu_count()
    if (n_cpus == 1):
        return get_ky_mat(hyps, training_data,
                          kernel, cutoffs, hyps_mask)

    # assume sigma_n is the final hyperparameter
    number_of_hyps = len(hyps)
    sigma_n = hyps[-1]
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask.keys()):
            if (hyps_mask['train_noise'] is False):
                sigma_n = hyps_mask['original'][-1]

    # initialize matrices

    size = len(training_data)
    size3 = 3*len(training_data)
    k_mat_slice = []
    k_mat = np.zeros([size3, size3])
    with mp.Pool(processes=n_cpus) as pool:

        ns = int(math.ceil(size/nsample))
        nproc = ns*(ns+1)//2
        if (nproc < n_cpus):
            nsample = int(math.ceil(size/int(np.sqrt(n_cpus*2))))
            ns = int(math.ceil(size/nsample))

        block_id = []
        nbatch = 0
        for ibatch1 in range(ns):
            s1 = int(nsample*ibatch1)
            e1 = int(np.min([s1 + nsample, size]))
            for ibatch2 in range(ibatch1, ns):
                s2 = int(nsample*ibatch2)
                e2 = int(np.min([s2 + nsample, size]))
                block_id += [(s1, e1, s2, e2)]
                nbatch += 1

        k_mat_slice = []
        count = 0
        base = 0
        time0 = time.time()
        for ibatch in range(nbatch):
            s1, e1, s2, e2 = block_id[ibatch]
            t1 = training_data[s1:e1]
            t2 = training_data[s2:e2]
            k_mat_slice.append(pool.apply_async(
                                      get_ky_mat_pack,
                                      args=(hyps,
                                        t1, t2, bool(s1==s2),
                                        kernel, cutoffs,
                                        hyps_mask)))
            count += 1
            if (count >= n_cpus*3):
                for iget in range(base, count+base):
                    s1, e1, s2, e2 = block_id[iget]
                    k_mat_block = k_mat_slice[iget-base].get()
                    k_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
                    if (s1 != s2):
                        k_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
                if (size>5000):
                    print("computed block", base, base+count, nbatch, time.time()-time0)
                    time0 = time.time()
                k_mat_slice = []
                base = ibatch+1
                count = 0
        if (count>0):
            if (size>5000):
                print("computed block", base, base+count, nbatch, time.time()-time0)
                time0 = time.time()
            for iget in range(base, nbatch):
                s1, e1, s2, e2 = block_id[iget]
                k_mat_block = k_mat_slice[iget-base].get()
                k_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
                if (s1 != s2):
                    k_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
            del k_mat_slice
        pool.close()
        pool.join()

    # matrix manipulation
    ky_mat = k_mat
    del k_mat_block
    ky_mat += sigma_n ** 2 * np.eye(size3)

    return ky_mat


def get_like_from_ky_mat(ky_mat, training_labels_np):
    """ compute the likelihood from the covariance matrix

    :param ky_mat: the covariance matrix
    :param training_labels_np: the numpy array of forces
    :type training_labels_np: np.array

    :return: float, likelihood
    """
    # catch linear algebra errors
    try:
        ky_mat_inv = np.linalg.inv(ky_mat)
        l_mat = np.linalg.cholesky(ky_mat)
    except:
        return -1e8

#######################################
##### KY MATRIX FUNCTIONS and gradients
#######################################


def get_ky_and_hyp(hyps: np.ndarray, training_data: list,
                   kernel_grad, cutoffs=None, hyps_mask=None):

    """
    computes ky matrix and its derivative to hyper-parameter
    If the cpu set up is None, it uses as much as posible cpus

    :param hyps: list of hyper-parameters
    :param training_data: list of atomic envirionments
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: hyp_mat, ky_mat
    """

    hyp_mat0, k_mat = get_ky_and_hyp_pack(hyps, training_data,
                                          training_data, True,
                                          kernel_grad,
                                          cutoffs=cutoffs,
                                          hyps_mask=hyps_mask)


    # obtain noise parameter
    train_noise = True
    sigma_n = hyps[-1]
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask):
            if (hyps_mask['train_noise'] is False):
                train_noise = False
                sigma_n = hyps_mask['original'][-1]
    # add gradient of noise variance
    size3 = 3*len(training_data)
    if (train_noise):
        sigma_mat = np.eye(size3) * 2 * sigma_n
        hyp_mat = np.zeros([hyp_mat0.shape[0]+1,
                            hyp_mat0.shape[1],
                            hyp_mat0.shape[2]])
        hyp_mat[-1, :, :] = sigma_mat
        hyp_mat[:-1, :, :] = hyp_mat0
    else:
        hyp_mat = hyp_mat0

    # matrix manipulation
    ky_mat = k_mat + sigma_n ** 2 * np.eye(size3)

    return hyp_mat, ky_mat


def get_ky_and_hyp_pack(hyps: np.ndarray, training_data1: list,
                   training_data2: list, same: bool,
                   kernel_grad, cutoffs=None, hyps_mask=None):
    """
    computes a block of ky matrix and its derivative to hyper-parameter
    If the cpu set up is None, it uses as much as posible cpus

    :param hyps: list of hyper-parameters
    :param training_data1: row elements
    :param training_data2: column elements
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: hyp_mat, ky_mat
    """

    # assume sigma_n is the final hyperparameter
    number_of_hyps = len(hyps)
    non_noise_hyps = len(hyps)-1
    train_noise = True
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask.keys()):
            if (hyps_mask['train_noise'] is False):
                train_noise = False
                non_noise_hyps = len(hyps)

    # initialize matrices
    size1 = len(training_data1) * 3
    size2 = len(training_data2) * 3
    k_mat = np.zeros([size1, size2])
    hyp_mat = np.zeros([non_noise_hyps, size1, size2])

    ds = [1, 2, 3]

    # calculate elements
    for m_index in range(size1):
        x_1 = training_data1[int(math.floor(m_index / 3))]
        d_1 = ds[m_index % 3]

        if (same):
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data2[int(math.floor(n_index / 3))]
            d_2 = ds[n_index % 3]

            # calculate kernel and gradient
            if (hyps_mask is not None):
                cov = kernel_grad(x_1, x_2, d_1, d_2, hyps,
                                  cutoffs=cutoffs, hyps_mask=hyps_mask)
            else:
                cov = kernel_grad(x_1, x_2, d_1, d_2, hyps, cutoffs)

            # store kernel value
            k_mat[m_index, n_index] = cov[0]
            hyp_mat[:, m_index, n_index] = cov[1]

            if (same):
                k_mat[n_index, m_index] = cov[0]
                hyp_mat[:, n_index, m_index] = cov[1]

    return hyp_mat, k_mat


def get_ky_and_hyp_par(hyps: np.ndarray, training_data: list,
                       kernel_grad, cutoffs=None,
                       hyps_mask=None,
                       n_cpus=None, nsample=100):
    """
    parallel version of get_ky_and_hyp

    :param hyps: list of hyper-parameters
    :param training_data: list of atomic envirionments
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param nsample: the size of block for matrix to compute

    :return: hyp_mat, ky_mat
    """


    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return get_ky_and_hyp(hyps, training_data,
                              kernel_grad,
                              cutoffs, hyps_mask)

    # assume sigma_n is the final hyperparameter
    number_of_hyps = len(hyps)
    non_noise_hyps = len(hyps)-1
    sigma_n = hyps[-1]
    train_noise = True
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask.keys()):
            if (hyps_mask['train_noise'] is False):
                sigma_n = hyps_mask['original'][-1]
                train_noise = False
                non_noise_hyps = len(hyps)

    # initialize matrices
    size = len(training_data)
    size3 = size*3
    k_mat = np.zeros([size3, size3])
    hyp_mat0 = np.zeros([non_noise_hyps, size3, size3])

    with mp.Pool(processes=n_cpus) as pool:

        ns = int(math.ceil(size/nsample))
        nproc = ns*(ns+1)//2
        if (nproc < n_cpus):
            nsample = int(math.ceil(size/np.sqrt(n_cpus*2)))
            ns = int(math.ceil(size/nsample))

        block_id = []
        nbatch = 0
        for ibatch1 in range(ns):
            s1 = int(nsample*ibatch1)
            e1 = int(np.min([s1 + nsample, size]))
            for ibatch2 in range(ibatch1, ns):
                s2 = int(nsample*ibatch2)
                e2 = int(np.min([s2 + nsample, size]))
                block_id += [(s1, e1, s2, e2)]
                nbatch += 1

        count = 0
        base = 0
        mat_slice = []
        for ibatch in range(nbatch):
            s1, e1, s2, e2 = block_id[ibatch]
            if (size>5000):
                print("sending", s1, e1, s2, e2)
            t1 = training_data[s1:e1]
            t2 = training_data[s2:e2]
            mat_slice.append(pool.apply_async(
                                    get_ky_and_hyp_pack,
                                    args=(
                                        hyps, t1, t2,
                                        bool(s1==s2),
                                        kernel_grad,
                                        cutoffs, hyps_mask)))

            count += 1
            if (count > n_cpus*3):
                for iget in range(base, count+base):
                    s1, e1, s2, e2 = block_id[iget]
                    h_mat_block, k_mat_block = mat_slice[iget-base].get()
                    k_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
                    hyp_mat0[:, s1*3:e1*3, s2*3:e2*3] = h_mat_block
                    if (s1 != s2):
                        k_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
                        for idx in range(hyp_mat0.shape[0]):
                            hyp_mat0[idx, s2*3:e2*3, s1*3:e1*3] = h_mat_block[idx].T
                    if (size>5000):
                        print("computed block", base, base+count)
                mat_slice = []
                base = ibatch+1
                count = 0
        if (count>0):
            for iget in range(base, nbatch):
                s1, e1, s2, e2 = block_id[iget]
                h_mat_block, k_mat_block = mat_slice[iget-base].get()
                k_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
                hyp_mat0[:, s1*3:e1*3, s2*3:e2*3] = h_mat_block
                if (s1 != s2):
                    k_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
                    for idx in range(hyp_mat0.shape[0]):
                        hyp_mat0[idx, s2*3:e2*3, s1*3:e1*3] = h_mat_block[idx].T
            del mat_slice
        pool.close()
        pool.join()

    # add gradient of noise variance
    if (train_noise):
        hyp_mat = np.zeros([hyp_mat0.shape[0]+1,
                            hyp_mat0.shape[1],
                            hyp_mat0.shape[2]])
        hyp_mat[:-1, :, :] = hyp_mat0
        hyp_mat[-1, :, :] = np.eye(size3) * 2 * sigma_n
    else:
        hyp_mat = hyp_mat0

    # matrix manipulation
    ky_mat = k_mat + sigma_n ** 2 * np.eye(size3)

    return hyp_mat, ky_mat


def get_neg_likelihood(hyps: np.ndarray, training_data: list,
                       training_labels_np: np.ndarray,
                       kernel: Callable, output = None,
                       cutoffs=None, hyps_mask=None,
                       n_cpus=None, nsample=100):
    """compute the negative log likelihood

    :param hyps: list of hyper-parameters
    :type hyps: np.ndarray
    :param training_data: Set of atomic environments to compare against
    :type training_data: list of AtomicEnvironment objects
    :param training_labels_np: forces
    :type training_labels_np: np.array
    :param kernel: function object of the kernel
    :param output: Output object for dumping every hyper-parameter
                   sets computed
    :type output: class Output
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param nsample: the size of block for matrix to compute

    :return: float
    """


    if output is not None:
        ostring="hyps:"
        for hyp in hyps:
            ostring+=f" {hyp}"
        ostring+="\n"
        output.write_to_log(ostring, name="hyps")

    time0 = time.time()
    ky_mat = \
        get_ky_mat_par(hyps, training_data, kernel,
                       cutoffs=cutoffs, hyps_mask=hyps_mask,
                       n_cpus=n_cpus, nsample=nsample)

    output.write_to_log(f"get_key_mat {time.time()-time0}\n", name="hyps")

    time0 = time.time()

    like = get_like_from_ky_mat(ky_mat, training_labels_np)

    output.write_to_log(f"get_like_from_ky_mat {time.time()-time0}\n",
                        name="hyps")

    if output is not None:
        output.write_to_log('like: ' + str(like)+'\n', name="hyps")

    return -like


def get_neg_like_grad(hyps: np.ndarray, training_data: list,
                      training_labels_np: np.ndarray,
                      kernel_grad, output = None,
                      cutoffs=None, hyps_mask=None,
                      n_cpus=None, nsample=100,
                      print_progress=False):
    """compute the log likelihood and its gradients

    :param hyps: list of hyper-parameters
    :type hyps: np.ndarray
    :param training_data: Set of atomic environments to compare against
    :type training_data: list of AtomicEnvironment objects
    :param training_labels_np: forces
    :type training_labels_np: np.array
    :param kernel_grad: function object of the kernel gradient
    :param output: Output object for dumping every hyper-parameter
                   sets computed
    :type output: class Output
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param nsample: the size of block for matrix to compute

    :return: float, np.array
    """

    time0 = time.time()
    if output is not None:
        ostring = "hyps:"
        for hyp in hyps:
            ostring +=  f" {hyp}"
        ostring += "\n"
        output.write_to_log(ostring, name="hyps")

    hyp_mat, ky_mat = \
        get_ky_and_hyp_par(hyps,
                           training_data,
                           kernel_grad,
                           cutoffs=cutoffs,
                           hyps_mask=hyps_mask,
                           n_cpus=n_cpus, nsample=nsample)

    if output is not None:
        output.write_to_log(f"get_ky_and_hyp {time.time()-time0}\n",
                            name="hyps")

    time0 = time.time()

    like, like_grad = \
        get_like_grad_from_mats(ky_mat, hyp_mat, training_labels_np)

    if output is not None:
        output.write_to_log(f"get_like_grad_from_mats {time.time()-time0}\n",
                            name="hyps")

    if output is not None:
        ostring = "like grad:"
        for lg in like_grad:
            ostring += f" {lg}"
        ostring += "\n"
        output.write_to_log(ostring, name="hyps")
        output.write_to_log('like: ' + str(like)+'\n', name="hyps")

    if print_progress:
        print('\nhyperparameters: '+str(hyps))
        print('likelihood: ' + str(like))
        print('likelihood gradient: ' + str(like_grad))

    return -like, -like_grad


def get_like_grad_from_mats(ky_mat, hyp_mat, training_labels_np):
    """compute the gradient of likelihood to hyper-parameters
    from covariance matrix and its gradient

    :param ky_mat: covariance matrix
    :type ky_mat: np.array
    :param hyp_mat: dky/d(hyper parameter) matrix
    :type hyp_mat: np.array
    :param training_labels_np: forces
    :type training_labels_np: np.array

    :return: float, list. the likelihood and its gradients
    """

    number_of_hyps = hyp_mat.shape[0]

    # catch linear algebra errors
    try:
        ky_mat_inv = np.linalg.inv(ky_mat)
        l_mat = np.linalg.cholesky(ky_mat)
    except:
        return -1e8, np.zeros(number_of_hyps)

    alpha = np.matmul(ky_mat_inv, training_labels_np)
    alpha_mat = np.matmul(alpha.reshape(alpha.shape[0], 1),
                          alpha.reshape(1, alpha.shape[0]))
    like_mat = alpha_mat - ky_mat_inv

    # calculate likelihood
    like = (-0.5 * np.matmul(training_labels_np, alpha) -
            np.sum(np.log(np.diagonal(l_mat))) -
            math.log(2 * np.pi) * ky_mat.shape[1] / 2)

    # calculate likelihood gradient
    like_grad = np.zeros(number_of_hyps)
    for n in range(number_of_hyps):
        like_grad[n] = 0.5 * \
                       np.trace(np.matmul(like_mat, hyp_mat[n, :, :]))

    return like, like_grad


def get_ky_mat_update_par(ky_mat_old, hyps: np.ndarray, training_data: list,
                      kernel, cutoffs=None, hyps_mask=None,
                      n_cpus=None, nsample=100):
    '''
    used for update_L_alpha, especially for parallelization
    parallelized for added atoms, for example, if add 10 atoms to the training
    set, the K matrix will add 10x3 columns and 10x3 rows, and the task will
    be distributed to 30 processors

    :param ky_mat_old: old covariance matrix
    :param hyps: list of hyper-parameters
    :param training_data: Set of atomic environments to compare against
    :param kernel:
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param nsample: the size of block for matrix to compute

    :return: updated covariance matrix

    '''

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return get_ky_mat_update(ky_mat_old, hyps, training_data,
                                 kernel, cutoffs, hyps_mask)

    # assume sigma_n is the final hyperparameter
    sigma_n = hyps[-1]
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask.keys()):
            if (hyps_mask['train_noise'] is False):
                sigma_n = hyps_mask['original'][-1]

    # initialize matrices
    old_size3 = ky_mat_old.shape[0]
    old_size = old_size3//3
    size = len(training_data)
    size3 = 3*len(training_data)
    ky_mat = np.zeros([size3, size3])
    ky_mat[:old_size3, :old_size3] = ky_mat_old
    ds = [1, 2, 3]
    with mp.Pool(processes=n_cpus) as pool:
        ns = int(math.ceil(size/nsample))
        nproc = (size3-old_size3)*(ns+old_size3)//2
        if (nproc < n_cpus):
            nsample = int(math.ceil(size/np.sqrt(n_cpus*2)))
            ns = int(math.ceil(size/nsample))

        ns_new = int(math.ceil((size-old_size)/nsample))
        old_ns = int(math.ceil(old_size/nsample))

        nbatch = 0
        block_id = []
        for ibatch1 in range(old_ns):
            s1 = int(nsample*ibatch1)
            e1 = int(np.min([s1 + nsample, old_size]))
            for ibatch2 in range(ns_new):
                s2 = int(nsample*ibatch2)+old_size
                e2 = int(np.min([s2 + nsample, size]))
                block_id += [(s1, e1, s2, e2)]
                nbatch += 1

        for ibatch1 in range(ns_new):
            s1 = int(nsample*ibatch1)+old_size
            e1 = int(np.min([s1 + nsample, size]))
            for ibatch2 in range(ns_new):
                s2 = int(nsample*ibatch2)+old_size
                e2 = int(np.min([s2 + nsample, size]))
                block_id += [(s1, e1, s2, e2)]
                nbatch += 1

        k_mat_slice = []
        count = 0
        base = 0
        time0 = time.time()
        for ibatch in range(nbatch):
            s1, e1, s2, e2 = block_id[ibatch]
            t1 = training_data[s1:e1]
            t2 = training_data[s2:e2]
            k_mat_slice += [pool.apply_async(
                                      get_ky_mat_pack,
                                      args=(hyps,
                                        t1, t2,
                                        bool(s1==s2),
                                        kernel, cutoffs, hyps_mask))]
            count += 1
            if (count >= n_cpus*3):
                for iget in range(base, count+base):
                    s1, e1, s2, e2 = block_id[iget]
                    k_mat_block = k_mat_slice[iget-base].get()
                    ky_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
                    if (s1 != s2):
                        ky_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
                if (size>5000):
                    print("computed block", base, base+count, nbatch, time.time()-time0)
                    time0 = time.time()
                k_mat_slice = []
                base = ibatch+1
                count = 0
        if (count>0):
            if (size>5000):
                print("computed block", base, base+count, nbatch, time.time()-time0)
                time0 = time.time()
            for iget in range(base, nbatch):
                s1, e1, s2, e2 = block_id[iget]
                k_mat_block = k_mat_slice[iget-base].get()
                ky_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
                if (s1 != s2):
                    ky_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
            del k_mat_slice
        pool.close()
        pool.join()

    # matrix manipulation
    ky_mat[old_size3:, old_size3:] += sigma_n ** 2 * np.eye(size3-old_size3)

    return ky_mat

def get_ky_mat_update(ky_mat_old, hyps: np.ndarray, training_data: list,
                      kernel, cutoffs=None, hyps_mask=None):
    '''
    used for update_L_alpha. if add 10 atoms to the training
    set, the K matrix will add 10x3 columns and 10x3 rows


    :param ky_mat_old: old covariance matrix
    :param hyps: list of hyper-parameters
    :param training_data: Set of atomic environments to compare against
    :param kernel:
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    '''
    n = ky_mat_old.shape[0]
    size = len(training_data)
    size3 = size*3
    m = size - n // 3  # number of new data added
    ky_mat = np.zeros((size3, size3))
    ky_mat[:n, :n] = ky_mat_old

    ds = [1, 2, 3]

    # calculate elements
    for m_index in range(size3):
        x_1 = training_data[int(math.floor(m_index / 3))]
        d_1 = ds[m_index % 3]
        low = int(np.max([m_index, n]))
        for n_index in range(low, size3):
            x_2 = training_data[int(math.floor(n_index / 3))]
            d_2 = ds[n_index % 3]

            # calculate kernel and gradient
            if (hyps_mask is not None):
                kern_curr = kernel(x_1, x_2, d_1, d_2, hyps,
                                   cutoffs, hyps_mask=hyps_mask)
            else:
                kern_curr = kernel(x_1, x_2, d_1, d_2, hyps,
                                   cutoffs)
            # store kernel value
            ky_mat[m_index, n_index] = kern_curr
            ky_mat[n_index, m_index] = kern_curr

    # matrix manipulation
    sigma_n = hyps[-1]
    if (hyps_mask is not None):
        if ('train_noise' in hyps_mask.keys()):
            if (hyps_mask['train_noise'] is False):
                sigma_n = hyps_mask['original'][-1]
    ky_mat[n:, n:] += sigma_n ** 2 * np.eye(size3-n)
    return ky_mat


def get_kernel_vector_unit(training_data, kernel, x,
                      d_1, hyps,
                      cutoffs=None, hyps_mask=None):
    """
    Compute kernel vector, comparing input environment to all environments
    in the GP's training set.
    :param training_data: Set of atomic environments to compare against
    :param kernel:
    :param x: data point to compare against kernel matrix
    :type x: AtomicEnvironment
    :param d_1: Cartesian component of force vector to get (1=x,2=y,3=z)
    :type d_1: int
    :param hyps: list of hyper-parameters
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: kernel vector
    :rtype: np.ndarray
    """


    ds = [1, 2, 3]
    size = len(training_data) * 3
    k_v = np.zeros(size, )

    for m_index in range(size):
        x_2 = training_data[int(math.floor(m_index / 3))]
        d_2 = ds[m_index % 3]
        if (hyps_mask is not None):
            k_v[m_index] = kernel(x, x_2, d_1, d_2,
                                  hyps, cutoffs, hyps_mask=hyps_mask)
        else:
            k_v[m_index] = kernel(x, x_2, d_1, d_2,
                                  hyps, cutoffs)
    return k_v

def get_kernel_vector_par(training_data, kernel,
                          x, d_1, hyps,
                          cutoffs=None, hyps_mask=None,
                          n_cpus=None, nsample=100):
    """
    Compute kernel vector, comparing input environment to all environments
    in the GP's training set.

    :param x: data point to compare against kernel matrix
    :type x: AtomicEnvironment
    :param d_1: Cartesian component of force vector to get (1=x,2=y,3=z)
    :type d_1: int
    :param hyps: list of hyper-parameters
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param nsample: the size of block for matrix to compute

    :return: kernel vector
    :rtype: np.ndarray
    """

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return get_kernel_vector(training_data, kernel,
                                 x, d_1, hyps,
                                 cutoffs, hyps_mask)

    with mp.Pool(processes=n_cpus) as pool:

        # sort of partition
        size = len(training_data)
        ns = int(math.ceil(size/nsample/n_cpus)*n_cpus)
        nsample = int(math.ceil(size/ns))
        ns = int(math.ceil(size/nsample))

        k12_slice = []
        for ibatch in range(ns):
            s = nsample*ibatch
            e = np.min([s + nsample, size])
            k12_slice.append(pool.apply_async(get_kernel_vector_unit,
                                              args=(training_data[s: e],
                                                    kernel, x, d_1, hyps,
                                                    cutoffs, hyps_mask)))

        size3 = size*3
        nsample3 = nsample*3
        k12_v = np.zeros(size3)
        for ibatch in range(ns):
            s = nsample3*ibatch
            e = np.min([s + nsample3, size3])
            k12_v[s:e] = k12_slice[ibatch].get()
        pool.close()
        pool.join()

    return k12_v

def get_kernel_vector(training_data, kernel,
                      x, d_1: int,
                      hyps, cutoffs=None, hyps_mask=None):
    """
    Compute kernel vector, comparing input environment to all environments
    in the GP's training set.

    :param x: data point to compare against kernel matrix
    :type x: AtomicEnvironment
    :param d_1: Cartesian component of force vector to get (1=x,2=y,3=z)
    :type d_1: int
    :param hyps: list of hyper-parameters
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    """

    ds = [1, 2, 3]
    size = len(training_data) * 3
    k_v = np.zeros(size, )

    for m_index in range(size):
        x_2 = training_data[int(math.floor(m_index / 3))]
        d_2 = ds[m_index % 3]
        if (hyps_mask is not None):
            k_v[m_index] = kernel(x, x_2, d_1, d_2,
                                  hyps, cutoffs,
                                  hyps_mask=hyps_mask)
        else:
            k_v[m_index] = kernel(x, x_2, d_1, d_2,
                                  hyps, cutoffs)

    return k_v



import sys
import numpy as np
import time
import copy
import multiprocessing as mp
import subprocess
from shutil import copyfile
from typing import List, Tuple, Union
from datetime import datetime

import flare.predict as predict
from flare import struc, gp, env, md
from flare.dft_interface import dft_software
from flare.output import Output
from flare.util import is_std_in_bound


class OTF:
    """Trains a Gaussian process force field on the fly during
        molecular dynamics.

    Args:
        dft_input (str): Input file.
        dt (float): MD timestep.
        number_of_steps (int): Number of timesteps in the training
            simulation.
        gp (gp.GaussianProcess): Initial GP model.
        dft_loc (str): Location of DFT executable.
        std_tolerance_factor (float, optional): Threshold that determines
            when DFT is called. Specifies a multiple of the current noise
            hyperparameter. If the epistemic uncertainty on a force
            component exceeds this value, DFT is called. Defaults to 1.
        prev_pos_init ([type], optional): Previous positions. Defaults
            to None.
        par (bool, optional): If True, force predictions are made in
            parallel. Defaults to False.
        skip (int, optional): Number of frames that are skipped when
            dumping to the output file. Defaults to 0.
        init_atoms (List[int], optional): List of atoms from the input
            structure whose local environments and force components are
            used to train the initial GP model. If None is specified, all
            atoms are used to train the initial GP. Defaults to None.
        calculate_energy (bool, optional): If True, the energy of each
            frame is calculated with the GP. Defaults to False.
        output_name (str, optional): Name of the output file. Defaults to
            'otf_run'.
        max_atoms_added (int, optional): Number of atoms added each time
            DFT is called. Defaults to 1.
        freeze_hyps (int, optional): Specifies the number of times the
            hyperparameters of the GP are optimized. After this many
            updates to the GP, the hyperparameters are frozen.
            Defaults to 10.
        rescale_steps (List[int], optional): List of frames for which the
            velocities of the atoms are rescaled. Defaults to [].
        rescale_temps (List[int], optional): List of rescaled temperatures.
            Defaults to [].
        force_source (Union[str, object], optional): DFT code used to calculate
            ab initio forces during training. A custom module can be used here
            in place of the DFT modules available in the FLARE package. The
            module must contain two functions: parse_dft_input, which takes a
            file name (in string format) as input and returns the positions,
            species, cell, and masses of a structure of atoms; and run_dft_par,
            which takes a number of DFT related inputs and returns the forces
            on all atoms.  Defaults to "qe".
        n_cpus (int, optional): Number of cpus used during training.
            Defaults to 1.
        npool (int, optional): Number of k-point pools for DFT
            calculations. Defaults to None.
        mpi (str, optional): Determines how mpi is called. Defaults to
            "srun".
        dft_kwargs ([type], optional): Additional arguments which are
            passed when DFT is called; keyword arguments vary based on the
            program (e.g. ESPRESSO vs. VASP). Defaults to None.
        store_dft_output (Tuple[Union[str,List[str]],str], optional):
            After DFT calculations are called, copy the file or files
            specified in the first element of the tuple to a directory
            specified as the second element of the tuple.
            Useful when DFT calculations are expensive and want to be kept
            for later use. The first element of the tuple can either be a
            single file name, or a list of several. Copied files will be
            prepended with the date and time with the format
            'Year.Month.Day:Hour:Minute:Second:'.
    """
    def __init__(self, dft_input: str, dt: float, number_of_steps: int,
                 gp: gp.GaussianProcess, dft_loc: str,
                 std_tolerance_factor: float = 1,
                 prev_pos_init: 'ndarray' = None, par: bool = False,
                 skip: int = 0, init_atoms: List[int] = None,
                 calculate_energy: bool = False, output_name: str = 'otf_run',
                 max_atoms_added: int = 1, freeze_hyps: int = 10,
                 rescale_steps: List[int] = [], rescale_temps: List[int] = [],
                 force_source: Union[str, object] = "qe", n_cpus: int = 1,
                 npool: int = None, mpi: str = "srun", dft_kwargs=None,
                 store_dft_output: Tuple[Union[str, List[str]], str] = None):

        self.dft_input = dft_input
        self.dt = dt
        self.number_of_steps = number_of_steps
        self.gp = gp
        self.dft_loc = dft_loc
        self.std_tolerance = std_tolerance_factor
        self.skip = skip
        self.dft_step = True
        self.freeze_hyps = freeze_hyps

        if isinstance(force_source, str):
            self.dft_module = dft_software[force_source]
        else:
            self.dft_module = force_source

        # parse input file
        positions, species, cell, masses = \
            self.dft_module.parse_dft_input(self.dft_input)

        self.structure = struc.Structure(cell=cell, species=species,
                                         positions=positions,
                                         mass_dict=masses,
                                         prev_positions=prev_pos_init,
                                         species_labels=species)

        self.noa = self.structure.positions.shape[0]
        self.atom_list = list(range(self.noa))
        self.curr_step = 0

        self.max_atoms_added = max_atoms_added

        # initialize local energies
        if calculate_energy:
            self.local_energies = np.zeros(self.noa)
        else:
            self.local_energies = None

        # set atom list for initial dft run
        if init_atoms is None:
            self.init_atoms = [int(n) for n in range(self.noa)]
        else:
            self.init_atoms = init_atoms

        self.dft_count = 0

        # set pred function
        if (par and gp.per_atom_par and gp.par) and not calculate_energy:
            self.pred_func = predict.predict_on_structure_par
        elif not calculate_energy:
            self.pred_func = predict.predict_on_structure
        elif (par and gp.per_atom_par and gp.par):
            self.pred_func = predict.predict_on_structure_par_en
        else:
            self.pred_func = predict.predict_on_structure_en
        self.par = par

        # set rescale attributes
        self.rescale_steps = rescale_steps
        self.rescale_temps = rescale_temps

        self.output = Output(output_name, always_flush=True)

        # set number of cpus and npool for DFT runs
        self.n_cpus = n_cpus
        self.npool = npool
        self.mpi = mpi

        self.dft_kwargs = dft_kwargs
        self.store_dft_output = store_dft_output

    def run(self):
        """
        Performs an on-the-fly training run.

        If OTF has store_dft_output set, then the specified DFT files will
        be copied with the current date and time prepended in the format
        'Year.Month.Day:Hour:Minute:Second:'.
        """

        self.output.write_header(self.gp.cutoffs, self.gp.kernel_name,
                                 self.gp.hyps, self.gp.algo,
                                 self.dt, self.number_of_steps,
                                 self.structure,
                                 self.std_tolerance)
        counter = 0
        self.start_time = time.time()

        while self.curr_step < self.number_of_steps:
            # run DFT and train initial model if first step and DFT is on
            if self.curr_step == 0 and self.std_tolerance != 0 and len(self.gp.training_data)==0:
                # call dft and update positions
                self.run_dft()
                dft_frcs = copy.deepcopy(self.structure.forces)
                new_pos = md.update_positions(self.dt, self.noa,
                                              self.structure)
                self.update_temperature(new_pos)
                self.record_state()

                # make initial gp model and predict forces
                self.update_gp(self.init_atoms, dft_frcs)
                if (self.dft_count-1) < self.freeze_hyps:
                    self.train_gp()

            # after step 1, try predicting with GP model
            else:
                self.gp.check_L_alpha()
                self.pred_func(self.structure, self.gp, self.n_cpus)
                self.dft_step = False
                new_pos = md.update_positions(self.dt, self.noa,
                                              self.structure)

                # get max uncertainty atoms
                std_in_bound, target_atoms = \
                    is_std_in_bound(self.std_tolerance,
                                    self.gp.hyps[-1], self.structure,
                                    self.max_atoms_added)

                if not std_in_bound:
                    # record GP forces
                    self.update_temperature(new_pos)
                    self.record_state()
                    gp_frcs = copy.deepcopy(self.structure.forces)

                    # run DFT and record forces
                    self.dft_step = True
                    self.run_dft()
                    dft_frcs = copy.deepcopy(self.structure.forces)
                    new_pos = md.update_positions(self.dt, self.noa,
                                                  self.structure)
                    self.update_temperature(new_pos)
                    self.record_state()

                    # compute mae and write to output
                    mae = np.mean(np.abs(gp_frcs - dft_frcs))
                    mac = np.mean(np.abs(dft_frcs))

                    self.output.write_to_log('\nmean absolute error:'
                                             ' %.4f eV/A \n' % mae)
                    self.output.write_to_log('mean absolute dft component:'
                                             ' %.4f eV/A \n' % mac)

                    # add max uncertainty atoms to training set
                    self.update_gp(target_atoms, dft_frcs)
                    if (self.dft_count-1) < self.freeze_hyps:
                        self.train_gp()

                    # Store DFT outputs in another folder if desired
                    # specified in self.store_dft_output
                    if self.store_dft_output is not None:
                        dest = self.store_dft_output[1]
                        target_files = self.store_dft_output[0]
                        now = datetime.now()
                        dt_string = now.strftime("%Y.%m.%d:%H:%M:%S:")
                        if isinstance(target_files, str):
                            to_copy = [target_files]
                        else:
                            to_copy = target_files
                        for file in to_copy:
                            copyfile(file, dest+'/'+dt_string+file)

            # write gp forces
            if counter >= self.skip and not self.dft_step:
                self.update_temperature(new_pos)
                self.record_state()
                counter = 0

            counter += 1
            self.update_positions(new_pos)
            self.curr_step += 1

        self.output.conclude_run()

    def run_dft(self):
        """Calculates DFT forces on atoms in the current structure.

        If OTF has store_dft_output set, then the specified DFT files will
        be copied with the current date and time prepended in the format
        'Year.Month.Day:Hour:Minute:Second:'.

        Calculates DFT forces on atoms in the current structure."""

        self.output.write_to_log('\nCalling DFT...\n')

        # calculate DFT forces
        forces = self.dft_module.run_dft_par(self.dft_input, self.structure,
                                             self.dft_loc,
                                             n_cpus=self.n_cpus,
                                             npool=self.npool,
                                             mpi=self.mpi,
                                             dft_kwargs=self.dft_kwargs)
        self.structure.forces = forces

        # write wall time of DFT calculation
        self.dft_count += 1
        self.output.write_to_log('DFT run complete.\n')
        time_curr = time.time() - self.start_time
        self.output.write_to_log('number of DFT calls: %i \n' % self.dft_count)
        self.output.write_to_log('wall time from start: %.2f s \n' % time_curr)

        # Store DFT outputs in another folder if desired
        # specified in self.store_dft_output
        if self.store_dft_output is not None:
            dest = self.store_dft_output[1]
            target_files = self.store_dft_output[0]
            now = datetime.now()
            dt_string = now.strftime("%Y.%m.%d:%H:%M:%S:")
            if isinstance(target_files, str):
                to_copy = [target_files]
            else:
                to_copy = target_files
            for file in to_copy:
                copyfile(file, dest+'/'+dt_string+file)

    def update_gp(self, train_atoms: List[int], dft_frcs: 'ndarray'):
        """
        Updates the current GP model.


        Args:
            train_atoms (List[int]): List of atoms whose local environments
                will be added to the training set.
            dft_frcs (np.ndarray): DFT forces on all atoms in the structure.
        """
        self.output.write_to_log('\nAdding atom {} to the training set.\n'
                                 .format(train_atoms))
        self.output.write_to_log('Uncertainty: {}.\n'
                                 .format(self.structure.stds[train_atoms[0]]))

        # update gp model
        self.gp.update_db(self.structure, dft_frcs,
                          custom_range=train_atoms)

        self.gp.set_L_alpha()

    def train_gp(self):
        """Optimizes the hyperparameters of the current GP model."""

        self.gp.train(self.output)
        self.output.write_hyps(self.gp.hyp_labels, self.gp.hyps,
                               self.start_time,
                               self.gp.likelihood, self.gp.likelihood_gradient)

    def update_positions(self, new_pos: 'ndarray'):
        """Performs a Verlet update of the atomic positions.

        Args:
            new_pos (np.ndarray): Positions of atoms in the next MD frame.
        """
        if self.curr_step in self.rescale_steps:
            rescale_ind = self.rescale_steps.index(self.curr_step)
            temp_fac = self.rescale_temps[rescale_ind] / self.temperature
            vel_fac = np.sqrt(temp_fac)
            self.structure.prev_positions = \
                new_pos - self.velocities * self.dt * vel_fac
        else:
            self.structure.prev_positions = self.structure.positions
        self.structure.positions = new_pos
        self.structure.wrap_positions()

    def update_temperature(self, new_pos: 'ndarray'):
        """Updates the instantaneous temperatures of the system.

        Args:
            new_pos (np.ndarray): Positions of atoms in the next MD frame.
        """
        KE, temperature, velocities = \
            md.calculate_temperature(new_pos, self.structure, self.dt,
                                     self.noa)
        self.KE = KE
        self.temperature = temperature
        self.velocities = velocities

    def record_state(self):
        self.output.write_md_config(self.dt, self.curr_step, self.structure,
                                    self.temperature, self.KE,
                                    self.local_energies, self.start_time,
                                    self.dft_step,
                                    self.velocities)



import time
import math
import pickle
import json

import numpy as np
import multiprocessing as mp

from copy import deepcopy
from typing import List, Callable
from scipy.linalg import solve_triangular
from scipy.optimize import minimize

from flare.env import AtomicEnvironment
from flare.struc import Structure
from flare.gp_algebra import get_neg_likelihood, \
                             get_like_from_ky_mat
from flare.gp_algebra import get_kernel_vector_par
from flare.gp_algebra import get_ky_mat_par
from flare.gp_algebra import get_ky_mat_update_par
from flare.gp_algebra import get_neg_like_grad
from flare.kernels.utils import str_to_kernels
from flare.util import NumpyEncoder
from flare.output import Output
import flare.cutoffs as cf


class GaussianProcess:
    """Gaussian process force field. Implementation is based on Algorithm 2.1
    (pg. 19) of "Gaussian Processes for Machine Learning" by Rasmussen and
    Williams.

    Args:
        kernel (Callable): Force/force kernel of the GP used to make force
            predictions.
        kernel_grad (Callable): Function that returns the gradient of the GP
            kernel with respect to the hyperparameters.
        hyps (np.ndarray): Hyperparameters of the GP.
        cutoffs (np.ndarray): Cutoffs of the GP kernel.
        hyp_labels (List, optional): List of hyperparameter labels. Defaults
            to None.
        energy_force_kernel (Callable, optional): Energy/force kernel of the
            GP used to make energy predictions. Defaults to None.
        energy_kernel (Callable, optional): Energy/energy kernel of the GP.
            Defaults to None.
        opt_algorithm (str, optional): Hyperparameter optimization algorithm.
            Defaults to 'L-BFGS-B'.
        maxiter (int, optional): Maximum number of iterations of the
            hyperparameter optimization algorithm. Defaults to 10.
        par (bool, optional): If True, the covariance matrix K of the GP is
            computed in parallel. Defaults to False.
        n_cpus (int, optional): Number of cpus used for parallel
            calculations. Defaults to 1.
        output (Output, optional): Output object used to dump hyperparameters
            during optimization. Defaults to None.
    """

    def __init__(self, kernel: Callable,
                 kernel_grad: Callable, hyps: 'ndarray',
                 cutoffs: 'ndarray',
                 hyp_labels: List = None,
                 energy_force_kernel: Callable = None,
                 energy_kernel: Callable = None,
                 opt_algorithm: str = 'L-BFGS-B',
                 maxiter: int = 10, par: bool = False,
                 per_atom_par: bool = True,
                 n_cpus: int = 1, nsample: int = 100,
                 output: Output = None,
                 multihyps: bool = False, hyps_mask: dict = None):
        """Initialize GP parameters and training data."""

        self.kernel = kernel
        self.kernel_grad = kernel_grad
        self.energy_kernel = energy_kernel
        self.energy_force_kernel = energy_force_kernel
        self.kernel_name = kernel.__name__
        self.hyps = hyps
        self.hyp_labels = hyp_labels
        self.cutoffs = cutoffs
        self.algo = opt_algorithm
        self.bounds = None

        self.training_data = []
        self.training_labels = []
        self.training_labels_np = np.empty(0, )
        self.maxiter = maxiter
        self.par = par
        self.per_atom_par = per_atom_par
        self.n_cpus = n_cpus
        self.nsample = nsample
        self.output = output

        # Parameters set during training
        self.ky_mat = None
        self.l_mat = None
        self.alpha = None
        self.ky_mat_inv = None
        self.likelihood = None
        self.likelihood_gradient = None

        if multihyps is True and hyps_mask is None:
            raise ValueError("Warning! Multihyperparameter mode enabled,"
                             "but no configuration hyperparameter mask was "
                             "passed. Did you mean to set multihyps to False?")
        elif multihyps is False and hyps_mask is not None:
            raise ValueError("Warning! Multihyperparameter mode disabled,"
                             "but a configuration hyperparameter mask was "
                             "passed. Did you mean to set multihyps to True?")
        self.hyps_mask = None
        if (isinstance(hyps_mask, dict) and multihyps is True):
            self.multihyps = True

            assert 'nspec' in hyps_mask.keys(), "nspec key missing in " \
                                                "hyps_mask dictionary"
            assert 'spec_mask' in hyps_mask.keys(), "spec_mask key missing " \
                                                    "in hyps_mask dicticnary"

            self.hyps_mask = deepcopy(hyps_mask)

            nspec = hyps_mask['nspec']

            if ('nbond' in hyps_mask.keys()):
                n2b = self.hyps_mask['nbond']
                if (n2b>0):
                    assert (np.max(hyps_mask['bond_mask']) < n2b)
                    assert len(hyps_mask['bond_mask']) == nspec**2, \
                            f"wrong dimension of bond_mask: " \
                            f" {len(hyps_mask['bond_mask']) != {nspec**2}}"
            else:
                n2b = 0

            if ('ntriplet' in hyps_mask.keys()):
                n3b = self.hyps_mask['ntriplet']
                if (n3b>0):
                    assert (np.max(hyps_mask['triplet_mask']) < n3b)
                    assert len(hyps_mask['triplet_mask']) == nspec**3, \
                            f"wrong dimension of triplet_mask" \
                            f"{len(hyps_mask['triplet_mask']) != {nspec**3}}"
            else:
                n3b = 0

            assert ((n2b+n3b)>0)

            if ('map' in hyps_mask.keys()):
                assert ('original' in hyps_mask.keys()), \
                        "original hyper parameters have to be defined"
                # Ensure typed correctly as numpy array
                self.hyps_mask['original'] = np.array(hyps_mask['original'])

                assert (n2b*2+n3b*2+1) == len(hyps_mask['original']) , \
                        "the hyperparmeter length is inconsistent with the mask"
                assert len(hyps_mask['map']) == len(hyps), \
                        "the hyperparmeter length is inconsistent with the mask"
                if ((len(hyps_mask['original'])-1) not in hyps_mask['map']):
                    assert hyps_mask['train_noise'] is False, \
                            "train_noise should be False when noise is not in hyps"
            else:
                assert hyps_mask['train_noise'] is True, \
                       "train_noise should be True when map is not used"
                assert (n2b*2+n3b*2+1) == len(hyps), \
                        "the hyperparmeter length is inconsistent with the mask"

            if ('bounds' in hyps_mask.keys()):
                self.bounds = deepcopy(hyps_mask['bounds'])
        else:
            self.multihyps = False
            self.hyps_mask = None

    # TODO unit test custom range
    def update_db(self, struc: Structure, forces: List,
                  custom_range: List[int] = ()):
        """Given a structure and forces, add local environments from the
        structure to the training set of the GP.

        Args:
            struc (Structure): Input structure. Local environments of atoms
                in this structure will be added to the training set of the GP.

            forces (np.ndarray): Forces on atoms in the structure.

            custom_range (List[int]): Indices of atoms whose local
                environments will be added to the training set of the GP.
        """

        # By default, use all atoms in the structure
        noa = len(struc.positions)
        update_indices = custom_range or list(range(noa))

        for atom in update_indices:
            env_curr = AtomicEnvironment(struc, atom, self.cutoffs)
            forces_curr = np.array(forces[atom])

            self.training_data.append(env_curr)
            self.training_labels.append(forces_curr)

        # create numpy array of training labels
        self.training_labels_np = np.hstack(self.training_labels)

    def add_one_env(self, env: AtomicEnvironment,
                    force, train: bool = False, **kwargs):
        """Add a single local environment to the training set of the GP.

        Args:
            env (AtomicEnvironment): Local environment to be added to the
                training set of the GP.
            force (np.ndarray): Force on the central atom of the local
                environment in the form of a 3-component Numpy array
                containing the x, y, and z components.
            train (bool): If True, the GP is trained after the local
                environment is added.
        """
        self.training_data.append(env)
        self.training_labels.append(force)
        self.training_labels_np = np.hstack(self.training_labels)

        if train:
            self.train(**kwargs)

    def train(self, output=None, custom_bounds=None,
              grad_tol: float = 1e-4,
              x_tol: float = 1e-5,
              line_steps: int = 20,
              print_progress: bool = False):
        """Train Gaussian Process model on training data. Tunes the
        hyperparameters to maximize the likelihood, then computes L and alpha
        (related to the covariance matrix of the training set).

        Args:
            output (Output): Output object specifying where to write the
                progress of the optimization.
            custom_bounds (np.ndarray): Custom bounds on the hyperparameters.
            grad_tol (float): Tolerance of the hyperparameter gradient that
                determines when hyperparameter optimization is terminated.
            x_tol (float): Tolerance on the x values used to decide when
                Nelder-Mead hyperparameter optimization is terminated.
            line_steps (int): Maximum number of line steps for L-BFGS
                hyperparameter optimization.
        """

        x_0 = self.hyps

        args = (self.training_data, self.training_labels_np,
                self.kernel_grad, output,
                self.cutoffs, self.hyps_mask,
                self.n_cpus, self.nsample,
                print_progress)
        objective_func = get_neg_like_grad
        res = None

        if self.algo == 'L-BFGS-B':

            # bound signal noise below to avoid overfitting
            if (self.bounds is None):
                bounds = np.array([(1e-6, np.inf)] * len(x_0))
                bounds[-1, 0] = 1e-3
            else:
                bounds = self.bounds

            # Catch linear algebra errors and switch to BFGS if necessary
            try:
                res = minimize(get_neg_like_grad, x_0, args,
                               method='L-BFGS-B', jac=True, bounds=bounds,
                               options={'disp': False, 'gtol': grad_tol,
                                        'maxls': line_steps,
                                        'maxiter': self.maxiter})
            except:
                print("Warning! Algorithm for L-BFGS-B failed. Changing to "
                      "BFGS for remainder of run.")
                self.algo = 'BFGS'

        if custom_bounds is not None:
            res = minimize(get_neg_like_grad, x_0, args,
                           method='L-BFGS-B', jac=True, bounds=custom_bounds,
                           options={'disp': False, 'gtol': grad_tol,
                                    'maxls': line_steps,
                                    'maxiter': self.maxiter})

        elif self.algo == 'BFGS':
            res = minimize(get_neg_like_grad, x_0, args,
                           method='BFGS', jac=True,
                           options={'disp': False, 'gtol': grad_tol,
                                    'maxiter': self.maxiter})

        elif self.algo == 'nelder-mead':
            res = minimize(get_neg_likelihood, x_0, args,
                           method='nelder-mead',
                           options={'disp': False,
                                    'maxiter': self.maxiter,
                                    'xtol': x_tol})
        if res is None:
            raise RuntimeError("Optimization failed for some reason.")
        self.hyps = res.x
        self.set_L_alpha()
        self.likelihood = -res.fun
        self.likelihood_gradient = -res.jac
        return res

    def check_L_alpha(self):
        """
        Check that the alpha vector is up to date with the training set. If
        not, update_L_alpha is called.
        """

        # check that alpha is up to date with training set
        if self.alpha is None or 3 * len(self.training_data) != len(
                self.alpha):
            self.update_L_alpha()

    def predict(self, x_t: AtomicEnvironment, d: int) -> [float, float]:
        """
        Predict a force component of the central atom of a local environment.

        Args:
            x_t (AtomicEnvironment): Input local environment.
            d (int): Force component to be predicted (1 is x, 2 is y, and
                3 is z).

        Return:
            (float, float): Mean and epistemic variance of the prediction.
        """

        # Kernel vector allows for evaluation of atomic environments.
        if (self.par and not self.per_atom_par):
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        k_v = get_kernel_vector_par(self.training_data, self.kernel,
                                    x_t, d,
                                    self.hyps,
                                    cutoffs=self.cutoffs,
                                    hyps_mask=self.hyps_mask,
                                    n_cpus=self.n_cpus,
                                    nsample=self.nsample)

        # Guarantee that alpha is up to date with training set
        if self.alpha is None or\
                (3 * len(self.training_data) != len(self.alpha)):
            self.update_L_alpha()

        # get predictive mean
        pred_mean = np.matmul(k_v, self.alpha)

        # get predictive variance without cholesky (possibly faster)
        # pass args to kernel based on if mult. hyperparameters in use
        if (self.multihyps):
            self_kern = self.kernel(x_t, x_t, d, d, self.hyps,
                                    self.cutoffs, hyps_mask=self.hyps_mask)
        else:
            self_kern = self.kernel(x_t, x_t, d, d, self.hyps,
                                    self.cutoffs)

        pred_var = self_kern - \
            np.matmul(np.matmul(k_v, self.ky_mat_inv), k_v)

        return pred_mean, pred_var

    def predict_local_energy(self, x_t: AtomicEnvironment) -> float:
        """Predict the local energy of a local environment.

        Args:
            x_t (AtomicEnvironment): Input local environment.

        Return:
            float: Local energy predicted by the GP.
        """

        k_v = self.en_kern_vec(x_t)
        pred_mean = np.matmul(k_v, self.alpha)

        return pred_mean

    def predict_local_energy_and_var(self, x_t: AtomicEnvironment):
        """Predict the local energy of a local environment and its
        uncertainty.

        Args:
            x_t (AtomicEnvironment): Input local environment.

        Return:
            (float, float): Mean and predictive variance predicted by the GP.
        """

        # get kernel vector
        k_v = self.en_kern_vec(x_t)

        # get predictive mean
        pred_mean = np.matmul(k_v, self.alpha)

        # get predictive variance
        v_vec = solve_triangular(self.l_mat, k_v, lower=True)
        if (self.multihyps):
            self_kern = self.energy_kernel(x_t, x_t, self.hyps,
                                           self.cutoffs,
                                           hyps_mask=self.hyps_mask)
        else:
            self_kern = self.energy_kernel(x_t, x_t, self.hyps,
                                           self.cutoffs)
        pred_var = self_kern - np.matmul(v_vec, v_vec)

        return pred_mean, pred_var

    def get_kernel_vector(self, x: AtomicEnvironment,
                          d_1: int):
        """
        Compute kernel vector, comparing input environment to all environments
        in the GP's training set.

        Args:
            x (AtomicEnvironment): Local environment to compare against
                the training environments.
            d_1 (int): Cartesian component of the kernel (1=x, 2=y, 3=z).

        Return:
            np.ndarray: Kernel vector.
        """

        ds = [1, 2, 3]
        size = len(self.training_data) * 3
        k_v = np.zeros(size, )

        if (self.multihyps):
            for m_index in range(size):
                x_2 = self.training_data[int(math.floor(m_index / 3))]
                d_2 = ds[m_index % 3]
                k_v[m_index] = self.kernel(x, x_2, d_1, d_2,
                                           self.hyps, self.cutoffs,
                                           hyps_mask=self.hyps_mask)
        else:
            for m_index in range(size):
                x_2 = self.training_data[int(math.floor(m_index / 3))]
                d_2 = ds[m_index % 3]
                k_v[m_index] = self.kernel(x, x_2, d_1, d_2,
                                           self.hyps, self.cutoffs)

        return k_v

    def en_kern_vec(self, x: AtomicEnvironment):
        """Compute the vector of energy/force kernels between an atomic
        environment and the environments in the training set.

        Args:
            x (AtomicEnvironment): Local environment to compare against
                the training environments.

        Return:
            np.ndarray: Kernel vector.
        """

        ds = [1, 2, 3]
        size = len(self.training_data) * 3
        k_v = np.zeros(size, )

        if (self.multihyps):
            for m_index in range(size):
                x_2 = self.training_data[int(math.floor(m_index / 3))]
                d_2 = ds[m_index % 3]
                k_v[m_index] = \
                    self.energy_force_kernel(x_2, x, d_2, self.hyps,
                                             self.cutoffs,
                                             hyps_mask=self.hyps_mask)
        else:
            for m_index in range(size):
                x_2 = self.training_data[int(math.floor(m_index / 3))]
                d_2 = ds[m_index % 3]
                k_v[m_index] = self.energy_force_kernel(x_2, x, d_2,
                                                        self.hyps,
                                                        self.cutoffs)

        return k_v

    def set_L_alpha(self):
        """
        Invert the covariance matrix, setting L (a lower triangular
        matrix s.t. L L^T = (K + sig_n^2 I)) and alpha, the inverse
        covariance matrix multiplied by the vector of training labels.
        The forces and variances are later obtained using alpha.
        """

        ky_mat = get_ky_mat_par(self.hyps,
                                self.training_data,
                                self.kernel,
                                cutoffs=self.cutoffs,
                                hyps_mask=self.hyps_mask,
                                n_cpus=self.n_cpus,
                                nsample=self.nsample)

        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.training_labels_np)

        self.ky_mat = ky_mat
        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv

        self.likelihood = \
            get_like_from_ky_mat(self.ky_mat, self.training_labels_np)

    def update_L_alpha(self):
        """
        Update the GP's L matrix and alpha vector without recalculating
        the entire covariance matrix K.
        """

        # Set L matrix and alpha if set_L_alpha has not been called yet
        if self.l_mat is None:
            self.set_L_alpha()
            return

        if (self.par and not self.per_atom_par):
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        ky_mat = get_ky_mat_update_par(self.ky_mat, self.hyps,
                                       self.training_data,
                                       self.kernel,
                                       cutoffs=self.cutoffs,
                                       hyps_mask=self.hyps_mask,
                                       n_cpus=n_cpus,
                                       nsample=self.nsample)

        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.training_labels_np)

        self.ky_mat = ky_mat
        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv

    def __str__(self):
        """String representation of the GP model."""

        thestr = "GaussianProcess Object\n"
        thestr += f'Kernel: {self.kernel_name}\n'
        thestr += f"Training points: {len(self.training_data)}\n"
        thestr += f'Cutoffs: {self.cutoffs}\n'
        thestr += f'Model Likelihood: {self.likelihood}\n'

        thestr += f'MultiHyps: {self.multihyps}\n'
        thestr += 'Hyperparameters: \n'
        if self.hyp_labels is None:
            # Put unlabeled hyperparameters on one line
            thestr = thestr[:-1]
            thestr += str(self.hyps) + '\n'
        else:
            for hyp, label in zip(self.hyps, self.hyp_labels):
                thestr += f"{label}: {hyp}\n"

        if (self.multihyps):
            nspec = self.hyps_mask['nspec']
            thestr += f'nspec: {nspec}\n'
            thestr += f'spec_mask: \n'
            thestr += str(self.hyps_mask['spec_mask']) + '\n'

            nbond = self.hyps_mask['nbond']
            thestr += f'nbond: {nbond}\n'
            if (nbond > 0):
                thestr +=  f'bond_mask: \n'
                thestr += str(self.hyps_mask['bond_mask']) + '\n'

            ntriplet = self.hyps_mask['ntriplet']
            thestr += f'ntriplet: {ntriplet}\n'
            if (ntriplet > 0):
                thestr += f'triplet_mask: \n'
                thestr += str(self.hyps_mask['triplet_mask']) + '\n'

        return thestr

    def as_dict(self):
        """Dictionary representation of the GP model."""

        out_dict = deepcopy(dict(vars(self)))

        out_dict['training_data'] = [env.as_dict() for env in
                                     self.training_data]
        # Remove the callables
        del out_dict['kernel']
        del out_dict['kernel_grad']

        return out_dict

    @staticmethod
    def from_dict(dictionary):
        """Create GP object from dictionary representation."""

        multihyps = dictionary.get('multihyps', False)

        force_kernel, grad = str_to_kernels(dictionary['kernel_name'],
                                            multihyps,
                                            include_grad=True)

        if dictionary['energy_kernel'] is not None:
            energy_kernel = str_to_kernel(dictionary['energy_kernel'],
                                          multihyps)
        else:
            energy_kernel = None

        if dictionary['energy_force_kernel'] is not None:
            energy_force_kernel = \
                str_to_kernel(dictionary['energy_force_kernel'],
                              multihyps)
        else:
            energy_force_kernel = None

        new_gp = GaussianProcess(kernel=force_kernel,
                                 kernel_grad=grad,
                                 energy_kernel=energy_kernel,
                                 energy_force_kernel=energy_force_kernel,
                                 cutoffs=np.array(dictionary['cutoffs']),
                                 hyps=np.array(dictionary['hyps']),
                                 hyp_labels=dictionary['hyp_labels'],
                                 par=dictionary['par'],
                                 per_atom_par=dictionary.get('per_atom_par',True),
                                 n_cpus=dictionary.get('n_cpus') or dictionary.get('no_cpus'),
                                 maxiter=dictionary['maxiter'],
                                 opt_algorithm=dictionary['algo'],
                                 multihyps=multihyps,
                                 hyps_mask=dictionary.get('hyps_mask',None)
                                 )

        new_gp.training_data = [AtomicEnvironment.from_dict(env) for env in
                                dictionary['training_data']]
        new_gp.training_labels = deepcopy(dictionary['training_labels'])
        new_gp.training_labels_np = deepcopy(dictionary['training_labels_np'])

        new_gp.likelihood = dictionary['likelihood']
        new_gp.likelihood_gradient = dictionary['likelihood_gradient']
        new_gp.training_labels_np = np.hstack(new_gp.training_labels)

        # Save time by attempting to load in computed attributes
        if (len(new_gp.training_data)>5000):
            new_gp.ky_mat = np.load(dictionary['ky_mat_file'])
            new_gp.compute_matrices()
        else:
            new_gp.ky_mat_inv = np.array(dictionary['ky_mat_inv']) \
                if dictionary.get('ky_mat_inv') is not None else None
            new_gp.ky_mat = np.array(dictionary['ky_mat']) \
                if dictionary.get('ky_mat') is not None else None
            new_gp.l_mat = np.array(dictionary['l_mat']) \
                if dictionary.get('l_mat') is not None else None
            new_gp.alpha = np.array(dictionary['alpha']) \
                if dictionary.get('alpha') is not None  else None
        return new_gp

    def compute_matrices(self):

        ky_mat = self.ky_mat
        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.training_labels_np)

        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv

    def write_model(self, name: str, format: str = 'json'):
        """
        Write model in a variety of formats to a file for later re-use.
        Args:
            name (str): Output name.
            format (str): Output format.
        """

        if (len(self.training_data)>5000):
            np.save(f"{name}_ky_mat.npy", self.ky_mat)
            self.ky_mat_file = f"{name}_ky_mat.npy"
            del self.ky_mat
            del self.l_mat
            del self.alpha
            del self.ky_mat_inv

        supported_formats = ['json', 'pickle', 'binary']

        if format.lower() == 'json':
            with open(f'{name}.json', 'w') as f:
                json.dump(self.as_dict(), f, cls=NumpyEncoder)

        elif format.lower() == 'pickle' or format.lower() == 'binary':
            with open(f'{name}.pickle', 'wb') as f:
                pickle.dump(self, f)

        else:
            raise ValueError("Output format not supported: try from "
                             "{}".format(supported_formats))

        if (len(self.training_data)>5000):
            self.ky_mat = np.load(f"{name}_ky_mat.npy")
            self.compute_matrices()


    @staticmethod
    def from_file(filename: str, format: str=''):
        """
        One-line convenience method to load a GP from a file stored using
        write_file

        Args:
            filename (str): path to GP model
            format (str): json or pickle if format is not in filename
        :return:
        """

        if '.json' in filename or 'json' in format:
            with open(filename, 'r') as f:
                return GaussianProcess.from_dict(json.loads(f.readline()))

        elif '.pickle' in filename or 'pickle' in format:
            with open(filename, 'rb') as f:
                return pickle.load(f)

        else:
            raise ValueError("Warning: Format unspecified or file is not "
                             ".json or .pickle format.")



'''
:class:`OTF` is the on-the-fly training module for ASE, WITHOUT molecular dynamics engine. 
It needs to be used adjointly with ASE MD engine. Please refer to our 
`OTF MD module <https://flare.readthedocs.io/en/latest/flare/ase/otf_md.html>`_ for the
complete training module with OTF and MD.
'''
import os
import sys
import inspect
from copy import deepcopy

from flare.struc import Structure
from flare.util import is_std_in_bound
from flare.mgp.utils import get_l_bound

import numpy as np
from ase.md.md import MolecularDynamics
from ase import units


class OTF:
    """
    OTF (on-the-fly) training with the ASE interface. 
    
    Note: Dft calculator is set outside of the otf module, and input as 
        dft_calc, so that different calculators can be used

    Args:
        dft_calc (ASE Calculator): the ASE DFT calculator (see ASE documentaion)
        dft_count (int): initial number of DFT calls
        std_tolerance_factor (float): the threshold of calling DFT = noise * 
            std_tolerance_factor
        init_atoms (list): the list of atoms in the first DFT call to add to
            the training set, since there's no uncertainty prediction initially
        calculate_energy (bool): if True, the energy will be calculated;
            otherwise, only forces will be predicted
        max_atoms_added (int): the maximal number of atoms to add to the 
            training set after each DFT calculation
        freeze_hyps (int or None): the hyperparameters will only be trained for
            the first `freeze_hyps` DFT calls, and will be fixed after that
        restart_from (str or None): the path of the directory that stores the
            training data from last OTF run, and this OTF will restart from it

    Other Parameters:
        use_mapping (bool): if True, the MGP will be used
        non_mapping_steps (list): a list of steps that MGP will not be 
            constructed and used
        l_bound (float): the lower bound of the interatomic distance, used for 
            MGP construction
        two_d (bool): used in the calculation of l_bound. If 2-D material is 
            considered, set to True, then the atomic environment construction 
            will only search the x & y periodic boundaries to save time
    """

    def __init__(self, 
            # on-the-fly parameters
            dft_calc=None, dft_count=None, std_tolerance_factor: float=1, 
            skip: int=0, init_atoms: list=[], calculate_energy=False, 
            max_atoms_added=1, freeze_hyps=1, restart_from=None,
            # mgp parameters
            use_mapping: bool=False, non_mapping_steps: list=[],
            l_bound: float=None, two_d: bool=False):

        # get all arguments as attributes 
        arg_dict = inspect.getargvalues(inspect.currentframe())[3]
        del arg_dict['self']
        self.__dict__.update(arg_dict)

        if dft_count is None:
            self.dft_count = 0
        self.noa = len(self.atoms.positions)

        # initialize local energies
        if calculate_energy:
            self.local_energies = np.zeros(self.noa)
        else:
            self.local_energies = None

        # initialize otf
        if init_atoms is None:
            self.init_atoms = [int(n) for n in range(self.noa)]

    def otf_run(self, steps, rescale_temp=[], rescale_steps=[]):
        """
        Use `otf_run` intead of `run` to perform a number of time steps.

        Args:
            steps (int): the number of time steps

        Other Parameters:
            rescale_temp (list): a list of temepratures that rescale the system
            rescale_steps (list): a list of step numbers that the temperature
                rescaling in `rescale_temp` is done

        Example:
            # rescale temperature to 500K and 1000K at the 100th and 200th step
            rescale_temp = [500, 1000]
            rescale_steps = [100, 200]
        """

        # observers
        for i, obs in enumerate(self.observers):
            if obs[0].__class__.__name__ == "OTFLogger":
                self.logger_ind = i
                break

        # restart from previous OTF training
        if self.restart_from is not None:
            self.restart()
            f = self.atoms.calc.results['forces']

        # initialize gp by a dft calculation
        if not self.atoms.calc.gp_model.training_data:
            self.dft_count = 0
            self.stds = np.zeros((self.noa, 3))
            dft_forces = self.call_DFT()
            f = dft_forces
   
            # update gp model
            curr_struc = Structure.from_ase_atoms(self.atoms)
            self.l_bound = get_l_bound(100, curr_struc, self.two_d)
            print('l_bound:', self.l_bound)

            self.atoms.calc.gp_model.update_db(curr_struc, dft_forces,
                           custom_range=self.init_atoms)

            # train calculator
            for atom in self.init_atoms:
                # the observers[0][0] is the logger
                self.observers[self.logger_ind][0].add_atom_info(atom, 
                    self.stds[atom])
            self.train()
            self.observers[self.logger_ind][0].write_wall_time()
  
        if self.md_engine == 'NPT':
            if not self.initialized:
                self.initialize()
            else:
                if self.have_the_atoms_been_changed():
                    raise NotImplementedError(
                        "You have modified the atoms since the last timestep.")

        step_0 = self.nsteps
        for i in range(step_0, steps):
            print('step:', i)
            self.atoms.calc.results = {} # clear the calculation from last step
            self.stds = np.zeros((self.noa, 3))

            # temperature rescaling
            if self.nsteps in rescale_steps:
                temp = rescale_temp[rescale_steps.index(self.nsteps)]
                curr_velocities = self.atoms.get_velocities()
                curr_temp = self.atoms.get_temperature()
                self.atoms.set_velocities(curr_velocities *\
                                          np.sqrt(temp/curr_temp))

            if self.md_engine == 'NPT':
                self.step()
            else:
                f = self.step(f)
            self.nsteps += 1
            self.stds = self.atoms.get_uncertainties(self.atoms)

            # figure out if std above the threshold
            self.call_observers() 
            curr_struc = Structure.from_ase_atoms(self.atoms)
            self.l_bound = get_l_bound(self.l_bound, curr_struc, self.two_d)
            print('l_bound:', self.l_bound)
            curr_struc.stds = np.copy(self.stds)
            noise = self.atoms.calc.gp_model.hyps[-1]
            self.std_in_bound, self.target_atoms = is_std_in_bound(\
                    noise, self.std_tolerance_factor, curr_struc, self.max_atoms_added)

            print('std in bound:', self.std_in_bound, self.target_atoms)
            #self.is_std_in_bound([])

            if not self.std_in_bound:
                # call dft/eam
                print('calling dft')
                dft_forces = self.call_DFT()

                # update gp
                print('updating gp')
                self.update_GP(dft_forces)

        self.observers[self.logger_ind][0].run_complete()

    
    def call_DFT(self):
        prev_calc = self.atoms.calc
        calc = deepcopy(self.dft_calc)
        self.atoms.set_calculator(calc)
        forces = self.atoms.get_forces()
        self.call_observers()
        self.atoms.set_calculator(prev_calc)
        self.dft_count += 1
        return forces

    def update_GP(self, dft_forces):
        atom_count = 0
        atom_list = []
        gp_model = self.atoms.calc.gp_model

        # build gp structure from atoms
        atom_struc = Structure.from_ase_atoms(self.atoms)

        while (not self.std_in_bound and atom_count <
               np.min([self.max_atoms_added, len(self.target_atoms)])):

            target_atom = self.target_atoms[atom_count]

            # update gp model
            gp_model.update_db(atom_struc, dft_forces,
                               custom_range=[target_atom])
    
            if gp_model.alpha is None:
                gp_model.set_L_alpha()
            else:
                gp_model.update_L_alpha()

            # atom_list.append(target_atom)
            ## force calculation needed before get_uncertainties
            # forces = self.atoms.calc.get_forces_gp(self.atoms) 
            # self.stds = self.atoms.get_uncertainties()

            # write added atom to the log file, 
            # refer to ase.optimize.optimize.Dynamics
            self.observers[self.logger_ind][0].add_atom_info(target_atom, 
                                               self.stds[target_atom])
           
            #self.is_std_in_bound(atom_list)
            atom_count += 1

        self.train()
        self.observers[self.logger_ind][0].added_atoms_dat.write('\n')
        self.observers[self.logger_ind][0].write_wall_time()

    def train(self, output=None, skip=False):
        calc = self.atoms.calc
        if (self.dft_count-1) < self.freeze_hyps:
            #TODO: add other args to train()
            calc.gp_model.train(output=output)
            self.observers[self.logger_ind][0].write_hyps(calc.gp_model.hyp_labels, 
                            calc.gp_model.hyps, calc.gp_model.likelihood, 
                            calc.gp_model.likelihood_gradient)
        else:
            #TODO: change to update_L_alpha()
            calc.gp_model.set_L_alpha()

        # build mgp
        if self.use_mapping:
            if self.get_time() in self.non_mapping_steps:
                skip = True

            calc.build_mgp(skip)

        np.save('ky_mat_inv', calc.gp_model.ky_mat_inv)
        np.save('alpha', calc.gp_model.alpha)

    def restart(self):
        # Recover atomic configuration: positions, velocities, forces
        positions, self.nsteps = self.read_frame('positions.xyz', -1)
        self.atoms.set_positions(positions)
        self.atoms.set_velocities(self.read_frame('velocities.dat', -1)[0])
        self.atoms.calc.results['forces'] = self.read_frame('forces.dat', -1)[0]
        print('Last frame recovered')

        # Recover training data set
        gp_model = self.atoms.calc.gp_model
        atoms = deepcopy(self.atoms)
        nat = len(self.atoms.positions)
        dft_positions = self.read_all_frames('dft_positions.xyz', nat)
        dft_forces = self.read_all_frames('dft_forces.dat', nat)
        added_atoms = self.read_all_frames('added_atoms.dat', 1, 1, 'int')
        for i, frame in enumerate(dft_positions):
            atoms.set_positions(frame)
            curr_struc = Structure.from_ase_atoms(atoms)
            gp_model.update_db(curr_struc, dft_forces[i], added_atoms[i])
        gp_model.set_L_alpha()
        print('GP training set ready')

        # Recover FLARE calculator
        gp_model.ky_mat_inv = np.load(self.restart_from+'/ky_mat_inv.npy')
        gp_model.alpha = np.load(self.restart_from+'/alpha.npy')
        if self.atoms.calc.use_mapping:
            for map_3 in self.atoms.calc.mgp_model.maps_3:
                map_3.load_grid = self.restart_from + '/'
            self.atoms.calc.build_mgp(skip=False)
        print('GP and MGP ready')

        self.l_bound = 10

    def read_all_frames(self, filename, nat, header=2, elem_type='xyz'):
        frames = []
        with open(self.restart_from+'/'+filename) as f:
            lines = f.readlines()
            frame_num = len(lines) // (nat+header)
            for i in range(frame_num):
                start = (nat+header) * i + header
                curr_frame = lines[start:start+nat]
                properties = []
                for line in curr_frame:
                    line = line.split()
                    if elem_type == 'xyz':
                        xyz = [float(l) for l in line[1:]]
                        properties.append(xyz)
                    elif elem_type == 'int':
                        properties = [int(l) for l in line]
                frames.append(properties)
        return np.array(frames)


    def read_frame(self, filename, frame_num):
        nat = len(self.atoms.positions)
        with open(self.restart_from+'/'+filename) as f:
            lines = f.readlines()
            if frame_num == -1: # read the last frame
                start_line = - (nat+2)
                frame = lines[start_line:]
            else:
                start_line = frame_num * (nat+2)
                end_line = (frame_num+1) * (nat+2)
                frame = f.lines[start_line:end_line]

            properties = []
            for line in frame[2:]:
                line = line.split()
                properties.append([float(d) for d in line[1:]])
        return np.array(properties), len(lines)//(nat+2)





"""
The :class:`Structure` object is a collection of atoms in a periodic box.
The mandatory inputs are the cell vectors of the box and the chemical species
and *Cartesian coordinates* of the atoms.
The atoms are automatically folded back into the primary cell, so the
input coordinates don't need to lie inside the box.
Energy, force, and stress information can be included which can then be
used to train ML models.
"""
import numpy as np
from flare.util import element_to_Z, NumpyEncoder
from json import dumps

from typing import List, Union, Any

try:
    # Used for to_pmg_structure method
    import pymatgen.core.structure as pmgstruc

    _pmg_present = True
except ImportError:
    _pmg_present = False


class Structure:
    """
    Contains information about a periodic structure of atoms, including the
    periodic cell boundaries, atomic species, and coordinates.

    *Note that input positions are assumed to be Cartesian.*

    :param cell: 3x3 array whose rows are the Bravais lattice vectors of the
        cell.
    :type cell: np.ndarray
    :param species: List of atomic species, which are represented either as
        integers or chemical symbols.
    :type species: List
    :param positions: Nx3 array of atomic coordinates.
    :type positions: np.ndarray
    :param mass_dict: Dictionary of atomic masses used in MD simulations.
    :type mass_dict: dict
    :param prev_positions: Nx3 array of previous atomic coordinates used in
        MD simulations.
    :type prev_positions: np.ndarray
    :param species_labels: List of chemical symbols. Used in the output file
        of on-the-fly runs.
    :type species_labels: List[str]
    :param stds: Uncertainty associated with forces
    :type stds: np.ndarray
    """

    def __init__(self, cell: 'ndarray', species: Union[List[str],
                                                          List[int]],
                 positions: 'ndarray',
                 mass_dict: dict = None,
                 prev_positions: 'ndarray' =None,
                 species_labels: List[str] = None,
                 forces=None,
                 stds=None):

        # Set up individual Bravais lattice vectors
        self.cell = np.array(cell)
        self.vec1 = self.cell[0, :]
        self.vec2 = self.cell[1, :]
        self.vec3 = self.cell[2, :]

        # get cell matrices for wrapping coordinates
        self.cell_transpose = self.cell.transpose()
        self.cell_transpose_inverse = np.linalg.inv(self.cell_transpose)
        self.cell_dot = self.get_cell_dot()
        self.cell_dot_inverse = np.linalg.inv(self.cell_dot)

        # set positions
        self.positions = np.array(positions)
        self.wrapped_positions = self.wrap_positions(in_place=False)

        # If species are strings, convert species to integers by atomic number
        if species_labels is None:
            self.species_labels = species
        else:
            self.species_labels = species_labels
        self.coded_species = np.array([element_to_Z(spec) for spec in species])
        self.nat = len(species)

        # Default: atoms have no velocity
        if prev_positions is None:
            self.prev_positions = np.copy(self.positions)
        else:
            assert len(positions) == len(prev_positions), 'Previous ' \
                                                          'positions and ' \
                                                          'positions are not' \
                                                          'same length'
            self.prev_positions = prev_positions

        self.energy = None
        self.stress = None

        if forces is not None:
            self.forces = np.array(forces)
        else:
            self.forces = np.zeros((len(positions), 3))

        if stds is not None:
            self.stds = np.array(stds)
        else:
            self.stds = np.zeros((len(positions), 3))

        self.mass_dict = mass_dict

    def get_cell_dot(self):
        """
        Compute 3x3 array of dot products of cell vectors used to
        fold atoms back to the unit cell.

        :return: 3x3 array of cell vector dot products.
        :rtype: np.ndarray
        """

        cell_dot = np.zeros((3, 3))

        for m in range(3):
            for n in range(3):
                cell_dot[m, n] = np.dot(self.cell[m], self.cell[n])

        return cell_dot

    @staticmethod
    def raw_to_relative(positions: 'ndarray', cell_transpose: 'ndarray',
                        cell_dot_inverse: 'ndarray')-> 'ndarray':
        """Convert Cartesian coordinates to relative (fractional) coordinates,
        expressed in terms of the cell vectors set in self.cell.

        :param positions: Cartesian coordinates.
        :type positions: np.ndarray
        :param cell_transpose: Transpose of the cell array.
        :type cell_transpose: np.ndarray
        :param cell_dot_inverse: Inverse of the array of dot products of
            cell vectors.
        :type cell_dot_inverse: np.ndarray
        :return: Relative positions.
        :rtype: np.ndarray
        """

        relative_positions = \
            np.matmul(np.matmul(positions, cell_transpose),
                      cell_dot_inverse)

        return relative_positions

    @staticmethod
    def relative_to_raw(relative_positions: 'ndarray',
                        cell_transpose_inverse: 'ndarray',
                        cell_dot: 'ndarray')-> 'ndarray':
        """Convert fractional coordinates to raw (Cartesian) coordinates.

        :param relative_positions: fractional coordinates.
        :type relative_positions: np.ndarray
        :param cell_transpose_inverse: Transpose of the cell array.
        :type cell_transpose_inverse: np.ndarray
        :param cell_dot: Dot products of cell vectors
        :type cell_dot: np.ndarray
        :return: Cartesian positions.
        :rtype: np.ndarray
        """

        return np.matmul(np.matmul(relative_positions, cell_dot),
                         cell_transpose_inverse)

    def wrap_positions(self, in_place: bool = True)-> 'ndarray':
        """
        Convenience function which folds atoms outside of the unit cell back
        into the unit cell. in_place flag controls if the wrapped positions
        are set in the class.

        :param in_place: If true, set the current structure 
		positions to be the wrapped positions.
        :return: Cartesian coordinates of positions all in unit cell
	:rtype: np.ndarray
        """
        rel_pos = \
            self.raw_to_relative(self.positions, self.cell_transpose,
                                 self.cell_dot_inverse)

        rel_wrap = rel_pos - np.floor(rel_pos)

        pos_wrap = self.relative_to_raw(rel_wrap, self.cell_transpose_inverse,
                                        self.cell_dot)

        if in_place:
            self.wrapped_positions = pos_wrap

        return pos_wrap

    def indices_of_specie(self, specie: Union[int, str]) -> List[int]:
        """
        Return the indices of a given species within atoms of the structure.

        :param specie: Element to target, can be string or integer
        :return: The indices in the structure at which this element occurs
        :rtype: List[str]
        """
        return [i for i, spec in enumerate(self.coded_species)
                if spec == specie]

    # TODO make more descriptive
    def __str__(self) -> str:
        """
        Simple descriptive string of structure.

        :return: One-line descriptor of number of atoms and species present.
        :rtype: str
        """

        return 'Structure with {} atoms of types {}'\
            .format(self.nat, set(self.species_labels))

    def __len__(self) -> int:
        """
        Returns number of atoms in structure.

        :return: number of atoms in structure.
        :rtype: int
        """
        return self.nat

    def as_dict(self) -> dict:
        """
        Returns structure as a dictionary; useful for serialization purposes.

        :return: Dictionary version of current structure
        :rtype: dict
        """
        return dict(vars(self))

    def as_str(self) -> str:
        """
        Returns string dictionary serialization cast as string.

        :return: output of as_dict method cast as string
        :rtype: str
        """
        return dumps(self.as_dict(), cls=NumpyEncoder)

    @staticmethod
    def from_dict(dictionary: dict)-> 'flare.struc.Structure':
        """
        Assembles a Structure object from a dictionary parameterizing one.

        :param dictionary: dict describing structure parameters.
        :return: FLARE structure assembled from dictionary
        """
        struc = Structure(cell=np.array(dictionary['cell']),
                          positions=np.array(dictionary['positions']),
                          species=dictionary['coded_species'],
                          forces=dictionary.get('forces'),
                          mass_dict=dictionary.get('mass_dict'),
                          species_labels=dictionary.get('species_labels'))

        struc.energy = dictionary.get('energy')
        struc.stress = dictionary.get('stress')
        struc.stds = np.array(dictionary.get('stds'))

        return struc

    @staticmethod
    def from_ase_atoms(atoms: 'ase.Atoms')-> 'flare.struc.Structure':
        """
        From an ASE Atoms object, return a FLARE structure

        :param atoms: ASE Atoms object
        :type atoms: ASE Atoms object
        :return: A FLARE structure from an ASE atoms object
        """
        struc = Structure(cell=np.array(atoms.cell),
                          positions=atoms.positions,
                          species=atoms.get_chemical_symbols())
        return struc

    def to_ase_atoms(self)-> 'ase.Atoms':
        from ase import Atoms
        return Atoms(self.species_labels,
                     positions=self.positions,
                     cell=self.cell)


    def to_pmg_structure(self):
        """
        Returns FLARE structure as a pymatgen structure.

        :return: Pymatgen structure corresponding to current FLARE structure
        """

        if not _pmg_present:
            raise ModuleNotFoundError("Pymatgen is not present. Please "
                                      "install Pymatgen and try again")

        site_properties = {'force:': self.forces, 'std': self.stds}

        return pmgstruc.Structure(lattice=self.cell,
                                  species=self.species_labels,
                                  coords=self.positions,
                                  coords_are_cartesian=True,
                                  site_properties=site_properties
                                  )

    @staticmethod
    def from_pmg_structure(structure: 'pymatgen Structure')-> \
            'flare Structure':
        """
        Returns Pymatgen structure as FLARE structure.

        :param structure: Pymatgen Structure
        :type structure: Pymatgen Structure
        :return: FLARE Structure
        """

        cell = structure.lattice.matrix.copy()
        species = [str(spec) for spec in structure.species]
        positions = structure.cart_coords.copy()

        new_struc = Structure(cell=cell, species=species,
                              positions=positions)

        site_props = structure.site_properties

        if 'force' in site_props.keys():
            forces = site_props['force']
            new_struc.forces = [np.array(force) for force in forces]

        if 'std' in site_props.keys():
            stds = site_props['std']
            new_struc.stds = [np.array(std) for std in stds]

        return new_struc


def get_unique_species(species: List[Any])-> (List, List[int]):
    """
    Returns a list of the unique species passed in, and a list of
    integers indexing them.

    :param species: Species to index uniquely
    :return: List of the unique species, and integer indexes
    """
    unique_species = []
    coded_species = []
    for spec in species:
        if spec in unique_species:
            coded_species.append(unique_species.index(spec))
        else:
            coded_species.append(len(unique_species))
            unique_species.append(spec)
    coded_species = np.array(coded_species)

    return unique_species, coded_species



"""
The cutoffs module gives a few different options for smoothly sending the GP
kernel to zero near the boundary of the cutoff sphere.
"""
from math import cos, sin, pi
from numba import njit


@njit
def hard_cutoff(r_cut: float, ri: float, ci: float):
    """A hard cutoff that assigns a value of 1 to all interatomic distances.

    Args:
        r_cut (float): Cutoff value (in angstrom).
        ri (float): Interatomic distance.
        ci (float): Cartesian coordinate divided by the distance.

    Returns:
        (float, float): Cutoff value and its derivative.
    """
    return 1, 0


@njit
def quadratic_cutoff(r_cut: float, ri: float, ci: float):
    """A quadratic cutoff that goes to zero smoothly at the cutoff boundary.

    Args:
        r_cut (float): Cutoff value (in angstrom).
        ri (float): Interatomic distance.
        ci (float): Cartesian coordinate divided by the distance.

    Returns:
        (float, float): Cutoff value and its derivative.
    """

    rdiff = r_cut - ri
    fi = rdiff * rdiff
    fdi = 2 * rdiff * ci

    return fi, fdi


@njit
def cubic_cutoff(r_cut: float, ri: float, ci: float):
    """A cubic cutoff that goes to zero smoothly at the cutoff boundary.

    Args:
        r_cut (float): Cutoff value (in angstrom).
        ri (float): Interatomic distance.
        ci (float): Cartesian coordinate divided by the distance.

    Returns:
        (float, float): Cutoff value and its derivative.
    """

    rdiff = r_cut - ri
    fi = rdiff * rdiff * rdiff
    fdi = 3 * rdiff * rdiff * ci

    return fi, fdi


@njit
def cosine_cutoff(r_cut: float, ri: float, ci: float, d: float = 1):
    """A cosine cutoff that returns 1 up to r_cut - d, and assigns a cosine
    envelope to values of r between r_cut - d and r_cut. Based on Eq. 24 of
    Albert P. Bartók and Gábor Csányi. "Gaussian approximation potentials: A
    brief tutorial introduction." International Journal of Quantum Chemistry
    115.16 (2015): 1051-1057.

    Args:
        r_cut (float): Cutoff value (in angstrom).
        ri (float): Interatomic distance.
        ci (float): Cartesian coordinate divided by the distance.

    Returns:
        (float, float): Cutoff value and its derivative.
    """

    if ri > r_cut - d:
        fi = (1/2) * (cos(pi * (ri - r_cut + d) / d) + 1)
        fdi = (pi/(2 * d)) * sin(pi * (r_cut - ri) / d) * ci
    else:
        fi = 1
        fdi = 0

    return fi, fdi



import numpy as np
import datetime
import time
import os

from ase.md import MDLogger
from ase import units

from flare.ase.calculator import FLARE_Calculator


class OTFLogger(MDLogger):

    def __init__(self, dyn, atoms, logfile, header=False, stress=False,
            peratom=False, mode="w", data_folder='otf_data', 
            data_in_logfile=False):

        super().__init__(dyn, atoms, logfile, header, stress,
                         peratom, mode)

        self.natoms = self.atoms.get_number_of_atoms()
        self.write_header_info()
        self.start_time = time.time()
        if data_folder not in os.listdir():
            os.system('mkdir '+data_folder)

        self.positions_xyz = open(data_folder+'/positions.xyz', mode=mode)
        self.velocities_dat = open(data_folder+'/velocities.dat', mode=mode)
        self.forces_dat = open(data_folder+'/forces.dat', mode=mode)
        self.uncertainties_dat = open(data_folder+'/uncertainties.dat', 
                                      mode=mode)
        self.dft_positions_xyz = open(data_folder+'/dft_positions.xyz', 
                                      mode=mode)
        self.dft_forces_dat = open(data_folder+'/dft_forces.dat', mode=mode)
        self.added_atoms_dat = open(data_folder+'/added_atoms.dat', mode=mode)

        self.traj_files = [self.positions_xyz, self.velocities_dat,
                self.forces_dat, self.uncertainties_dat]
        self.dft_data_files = [self.dft_positions_xyz, self.dft_forces_dat]
        self.data_in_logfile = data_in_logfile

    def write_header_info(self):
        gp_model = self.atoms.calc.gp_model
        self.logfile.write(str(datetime.datetime.now()))
        self.logfile.write('\nnumber of cpu cores: ')  # TODO
        self.logfile.write('\ncutoffs: '+str(gp_model.cutoffs))
        self.logfile.write('\nkernel: '+gp_model.kernel.__name__)
        self.logfile.write('\nnumber of hyperparameters: '+str(len(gp_model.hyps)))
        self.logfile.write('\nhyperparameters: '+str(gp_model.hyps))
        self.logfile.write('\nhyperparameter optimization algorithm: ' +
                           gp_model.algo)
        self.logfile.write('\nuncertainty tolerance: {} times noise'.format(
                           str(self.dyn.std_tolerance_factor)))
        self.logfile.write('\ntimestep (ps): {}'.format(self.dyn.dt/1000))
        self.logfile.write('\nnumber of frames: {}'.format(0))
        self.logfile.write('\nnumber of atoms: {}'.format(
                           len(self.atoms.positions)))
        self.logfile.write('\nsystem species: {}'.format(
                           self.atoms.get_chemical_symbols()))
        self.logfile.write('\nperiodic cell:\n'+str(np.array(self.atoms.cell)))
        self.logfile.write('\n')
        self.write_prev_positions()


    def write_hyps(self, hyp_labels, hyps, like, like_grad):
        self.logfile.write('\n\nGP hyperparameters: \n')
        for i, label in enumerate(hyp_labels):
            self.logfile.write('Hyp{} : {} = {}\n'.format(i, label, hyps[i]))

        self.logfile.write('likelihood: '+str(like)+'\n')
        self.logfile.write('likelihood gradient: '+str(like_grad))

    def write_wall_time(self):
        self.logfile.write('\nwall time from start: %.2f s \n'
                           % (time.time()-self.start_time))

    def write_prev_positions(self):
        v = self.atoms.get_velocities()
        pos = self.atoms.get_positions()
        dt = self.dyn.dt
        prev_pos = pos - v * dt
        species = self.atoms.get_chemical_symbols()
        nat = len(pos)
        self.logfile.write('previous positions (A):\n')
        template = '{} {:9f} {:9f} {:9f}\n'
        for n in range(nat):
            self.logfile.write(template.format(species[n], 
                prev_pos[n,0], prev_pos[n,1], prev_pos[n,2]))

    def __call__(self):
        self.write_logfile()
        self.write_datafiles()

    def write_datafiles(self):
        template = '{} {:9f} {:9f} {:9f}'
        steps = self.dyn.nsteps
        t = steps / 1000

        species = self.atoms.get_chemical_symbols()
        positions = self.atoms.get_positions()
        forces = self.atoms.get_forces()
        if type(self.atoms.calc) == FLARE_Calculator: 
            velocities = self.atoms.get_velocities()
            stds = self.atoms.get_uncertainties(self.atoms)
            data_files = self.traj_files
            data = [positions, velocities, forces, stds]
        else:
            data_files = self.dft_data_files
            data = [positions, forces]
            self.added_atoms_dat.write('Frame '+str(steps)+'\n')
         
        for ind, f in enumerate(data_files):
            f.write(str(self.natoms))
            f.write('\nFrame '+str(steps)+'\n')
            for atom in range(self.natoms):
                dat = template.format(species[atom],
                                      data[ind][atom][0], 
                                      data[ind][atom][1], 
                                      data[ind][atom][2])
                f.write(dat+'\n')
            f.flush()


    def write_logfile(self):
        self.logfile.write(50*'-')
        if self.dyn is not None:
            steps = self.dyn.nsteps
            t = steps / 1000
            if type(self.atoms.calc) != FLARE_Calculator: 
                self.logfile.write('\n*-Frame: '+str(steps))
            else:
                self.logfile.write('\n-Frame: '+str(steps))
            self.logfile.write('\nSimulation time: '+str(t)+' ps')
        self.logfile.write('\n')

        if self.data_in_logfile:
            self.write_data_to_logfile()

        # write energy, temperature info
        epot = self.atoms.get_potential_energy()
        ekin = self.atoms.get_kinetic_energy()
        temp = ekin / (1.5 * units.kB * self.natoms)
        if self.peratom:
            epot /= self.natoms
            ekin /= self.natoms
#        self.logfile.write('\ntotal energy: '+str(epot+ekin))
        self.logfile.write('\ntemperature: '+str(temp)+' K')
        self.logfile.write('\nkinetic energy: '+str(ekin)+' eV')
        self.write_wall_time()

        self.logfile.flush()
 
    def write_data_to_logfile(self):
        # add positions, forces and stds to be written
        species = self.atoms.get_chemical_symbols()
        positions = self.atoms.get_positions()
        forces = self.atoms.get_forces()
        velocities = self.atoms.get_velocities()
        if type(self.atoms.calc) == FLARE_Calculator: 
            stds = self.atoms.get_uncertainties(self.atoms)
            force_str = 'GP  Forces'
        else:
            stds = np.zeros(positions.shape)
            force_str = 'DFT Forces'
        self.logfile.write('Type'+(7*' ') + 'Positions'+(23*' ') +
                           force_str+(22*' ') + 'Uncertainties'+(22*' ') +
                           'Velocities\n')

        template = '{} {:9f} {:9f} {:9f}   {:9f} {:9f} {:9f}   '\
                   '{:9f} {:9f} {:9f}   {:9f} {:9f} {:9f}'
        for atom in range(len(positions)):
            dat = \
                template.format(species[atom], positions[atom][0], 
                                positions[atom][1], positions[atom][2], 
                                forces[atom][0], forces[atom][1], 
                                forces[atom][2], stds[atom][0],
                                stds[atom][1], stds[atom][2],
                                velocities[atom][0], velocities[atom][1],
                                velocities[atom][2])
            self.logfile.write(dat+'\n')

    def add_atom_info(self, target_atom, uncertainty):
        if not isinstance(target_atom, list):
            target_atom = [target_atom]
        self.logfile.write('\nAdding atom {} to the training set.\
                            \nUncertainty: {}.'.format(target_atom, uncertainty))
        self.added_atoms_dat.write(str(target_atom[0])+' ') # temporarily support 1 atom

    def write_mgp_train(self, mgp_model, train_time):
        train_size = len(mgp_model.GP.training_data)
        self.logfile.write('\ntraining set size: {}\n'.format(train_size))
        self.logfile.write('lower bound: {}\n'.format(mgp_model.l_bound))
        self.logfile.write('mgp l_bound: {}\n'.format(mgp_model.grid_params
                                                      ['bounds_2'][0, 0]))
        self.logfile.write('building mapping time: {}'.format(train_time))

    def run_complete(self):
        self.logfile.write('Run complete.')
        for f in self.traj_files:
            f.close()
        for f in self.dft_data_files:
            f.close()
